import os.path
import json
from pathlib import Path
import subprocess 
from typing import Any
from .external import batched,islice
import pandas as pd

def file_to_batches(fn_in: str, batchlength: int) -> list[str]:
    """Splits file into list of strings with every N lines as items"""
    with open(fn_in) as file:
        complete_linelist: list[str] = file.readlines()
        total_linecount: int = len(complete_linelist)
        batch_holder: list[str] = []
        batches = batched(complete_linelist,batchlength)
        for batch in batches: 
            sublist_linecount = len(batch)
            linecounter = 0
            query_input: str = ""
            for line in batch:
                if linecounter != sublist_linecount:       
                    query_input += line[:-1]+"\\n"
                    linecounter += 1
                else:
                    batch += line[-1]
            batch_holder.append(query_input)
    return batch_holder

def run_subprocess(command: str,**kwargs) -> None:
    """Run a specific command from the shell"""
    if "cwd" in kwargs:
        cwd = kwargs["cwd"]
        subprocess.run([command], cwd=cwd, shell=True, capture_output=True, text=True)
    else:
        subprocess.run([command], shell=True, capture_output=True, text=True)

def validate_dictkeys(to_check: dict,required_keys: list[str]):
    """Check if all items from a list are dictkeys"""
    return all(required_key in to_check for required_key in required_keys)

def json_file_to_dict(fn):
    """Converts a file in Json format to a dict"""
    with open(fn,'r') as file:
        content: dict = json.load(file)
    return content

def json_from_string_or_file(input_json) -> dict:
        """returns input json from file or dict as dict"""
        # self.paths,self.tools,self.other = workflowsettings.values()
        if isinstance(input_json,str):
            try:
                os.path.exists(input_json)
            except FileNotFoundError: 
                print(f"Settings file not found: {input_json}")
            with open(input_json,'r') as file:
                content: dict = json.load(file)
        elif isinstance(input_json,dict):
            content: dict = input_json
        else: 
            raise TypeError(f"json has to be file or dict, not {type(input_json)}")
        return content

def get_filelist_from_folder(folder: str) -> list:
    """Get all files from a folder as list"""
    p = Path(folder)
    filelist: list = [file.as_posix() for file in p.iterdir()]
    return filelist


# l_pos_fn, l_neg_fn = split_files_per_mode(data_folder)
#     for paramset in ["strictest","intermediate","lenient","manual"]:
#         write_mzbatch(l_files=l_pos_fn,mzbatch_fn=basis_mzbatch,metadata_fn=fn_metadata,paramset=paramset,output_folder=mzbatch_output_loc)

def query_df_value_by_other_col(query,df:pd.DataFrame,query_col,target_col) -> Any:
    """Checks if query col contains query and if so returns value from target col"""
    results = df.loc[:,[target_col]][df[query_col] == query]
    #df.loc[:,[query_col]][df  == query]
    return results.iat[0,0] if len(results) else "N/A"

def integrate_df_col_to_df(target:pd.DataFrame,t_vals,t_queries,source:pd.DataFrame,s_vals,s_queries) -> pd.DataFrame:
    """Integrate value column from source into target if query columns match"""
    target[t_vals] = target[t_queries].map(lambda t_query: 
                query_df_value_by_other_col(query=t_query,df=source,query_col=s_queries,target_col=s_vals))     
    return target

def integrate_df_cols_to_df(target_df:pd.DataFrame,source_df:pd.DataFrame,
                                target_integration_col: Any,source_integration_col: Any,
                                col_translator:dict[Any,Any]) -> pd.DataFrame:
    """Integrate values columns from source into target if query columns match for value columns in col_translator"""
    for old_name,new_name in col_translator.items():
        target_info = target_df,new_name, target_integration_col
        source_info = source_df.loc[:,[old_name,source_integration_col]],old_name,source_integration_col
        target_df = integrate_df_col_to_df(*target_info,*source_info)
    return target_df

def convert_missing(x):
    """Helper to convert missing value to the same format"""
    if (x=="nan") or (x.lower() == "n/a"):
        return "N/A"
    else:
        return x


import pandas as pd
import matplotlib.pyplot as plt
from typing import Any,Iterable

# def plot_classes(df_in,level):
#     df_in[f'cf_{level}']
#     df_superclass = df_in[f'cf_{level}']
#     #df_superclass = df_sdf_relevant.map(lambda x: x.split("<Superclass>\n")[1].split("\n")[0]) #Parse superclass
#     plastchem_superclasscounts = df_superclass.value_counts(normalize=True)
#     d_plastchem_superclasscounts = plastchem_superclasscounts.to_dict()
#     d_temp = {}
#     for i in d_plastchem_superclasscounts:
#         d_temp[i] = d_plastchem_superclasscounts[i]
#     d_plastchem_superclasscounts = d_temp
#     df_pcc = pd.DataFrame.from_dict(d_plastchem_superclasscounts,orient="index",columns=["count"])
#     data_superclasscounts = integrated_df[f'canopus:CF_{level}'].value_counts(normalize=True)
#     d_data_superclasscounts = data_superclasscounts.to_dict()
#     df_dsc = pd.DataFrame.from_dict(d_data_superclasscounts,orient="index",columns=["count"])
#     df_joined = df_dsc.join(df_pcc,lsuffix="_data",rsuffix="_plastchem", how="outer")
#     df_joined = df_joined.fillna(0)
#     ax = df_joined[['count_data','count_plastchem']].plot(kind='bar', title ="Chemical space", figsize=(15, 10), legend=True, fontsize=12)
#     ax.set_xlabel(level, fontsize=12)
#     fig_name=f"/lustre/BIF/nobackup/hendr218/temp/Plots/plastchem_data_{level}.png"
#     plt.savefig(fig_name)


def create_counts_df(df: pd.DataFrame,col_to_plot: str,col_to_filter:str=None,normalize:bool=True) -> pd.DataFrame:
    """Creates a dataframe for """
    if col_to_filter:
        df = df[col_to_plot][(df[col_to_filter] == True) & (df["is_blank"] == False)].value_counts(normalize=normalize)
    else: 
        df = df[col_to_plot][df["is_blank"] == False].value_counts(normalize=normalize)
    df=df.to_dict()
    df = pd.DataFrame.from_dict(df,orient="index",columns=[col_to_plot])
    return df

#fig_name=f"/lustre/BIF/nobackup/hendr218/Data/cf_only_{vis_col}_by_plastic_type.png"

# def plot_count_cols(fig_name:str):
#     for vis_col in ["CF:subclass", "CF:class", "CF:superclass"]:
#         nw = onw[onw[vis_col] != "N/A"]
#         PE_PET = create_vc_df(nw,vis_col,col_to_filter="PE_PET")
#         PE_PA = create_vc_df(nw,vis_col,col_to_filter="PE_PA")
#         PE = create_vc_df(nw,vis_col,col_to_filter="PE")
#         ALL = create_vc_df(nw,vis_col)

#         df_joined = PE.join(PE_PET,lsuffix="_PE",rsuffix="_PET", how="outer")
#         df_joined2 = ALL.join(PE_PA,lsuffix="_ALL",rsuffix="_PA", how="outer")
#         df_joined = df_joined2.join(df_joined,how="outer").fillna(0)
#         df_joined = df_joined.apply(lambda x: x*100)
#         df_joined = df_joined.rename(columns=lambda x: x.split("_")[-1])
#         if vis_col == "subclass":
#             df_joined = df_joined[df_joined.apply(lambda row: any([i > 1  for i in row]),axis=1)] 


# def join_counts(*dataframes: Iterable[pd.DataFrame]):
#     if len(args) == 1:
#         return args
#     dfs = [dataframe for dataframe in dataframes]
#     df_joined = dfs[0]
#     dfs.pop[0]
#     for i,df in dfs.enumerate():
#         df_joined = df_joined.join(df,lsuffix=f"_{i-1}",rsuffix=f"_{i}",how="outer")
#     df_joined.fillna(0)


# def df_to_barplot(df:pd.DataFrame,columns:list[str],title: str,ytitle:str,fn: str) -> None:
#     ax = df[columns].plot(kind='bar', title=title, figsize=(15, 10), legend=True, fontsize=12)
#     ax.set_ylabel(ytitle, fontsize=12)
#     plt.tight_layout()
#     plt.savefig(fn)


import os
import numpy as np
import json
from typing import Any

from matchms.importing import load_from_mgf
from matplotlib import pyplot as plt
from matchms.filtering import remove_peaks_around_precursor_mz, remove_peaks_outside_top_k,select_by_mz
from matchms import calculate_scores
import matchms.Scores
from matchms.similarity import ModifiedCosine
import networkx as nx
# import matchmsextras.networking as net
from matchms.networking import SimilarityNetwork
import pandas as pd
from typing import Any,Iterable
import rdkit.Chem.PandasTools

try:
    import MS2LDA
except ImportError:
     MS2LDA = None

def network_to_edgelist_and_nodes_df(graph: nx.Graph) -> tuple[pd.DataFrame, pd.DataFrame]:
    """splits a networkx graph into an edge and nodes dataframe"""
    nodes = graph.nodes()
    for node in nodes:
      nodes[node]["placeholder"] = "placeholder"
    #    networkdf = pd.DataFrame.from_dict(dict(graph.nodes(data=True)), orient='index')  
    networkdf = pd.DataFrame.from_dict(nodes,orient='index')      

    edgelist = nx.to_pandas_edgelist(graph)
    return networkdf, edgelist


def integrate_sirius_to_graph(graph: nx.Graph,translators: tuple[dict,dict,dict],outputs: tuple[str,str,str]) -> nx.Graph:
    """Merge results from sirius tools to graph"""
    sirius_translator,fingerid_translator,canopus_translator = translators
    sirius_file,fingerid_file,canopus_file = outputs
    sirius_df = pd.read_csv(sirius_file, sep='\t')
    fingerid_df = pd.read_csv(fingerid_file, sep="\t")
    canopus_df = pd.read_csv(canopus_file, sep="\t")

    graph = df_to_graph(graph,canopus_df,canopus_translator)
    graph = df_to_graph(graph,sirius_df,sirius_translator)
    graph = df_to_graph(graph,fingerid_df,fingerid_translator)

    return graph

def df_to_graph(graph: nx.Graph,to_merge: pd.DataFrame,translator: dict[str,str]) -> nx.Graph:
    """Adds columns from dataframe to nodes, merging on "scan" columns from DF"""
    for result in to_merge.to_dict(orient="records"):
        scan = str(result["mappingFeatureId"])
        if scan in graph:
            for oldname,newname in translator.items():
                graph.nodes[scan][newname] = result[oldname]
    return graph

def integrate_ms2lda_to_df(dataframe: pd.DataFrame, dataset: str,output_folder: str,output_database: str) -> pd.DataFrame:
    """Get motifs per feature and use feature ID to merge motifs to network"""
    motifDB_query = "QUERY scaninfo(MS2DATA)"
    screening_hits = MS2LDA.screen_spectra(motifDB=output_database, dataset=dataset, motifDB_query=motifDB_query, output_folder=output_folder)
    with open(dataset) as file:
        A = file.read()
        feature_ids = [entry.split('\n')[1][11:] for entry in A.split("BEGIN IONS")[1:]]
        translation_dict = {}
        for i,j in enumerate(feature_ids):
            translation_dict[f'spec_{i}'] = j
        screening_hits["scan"] = screening_hits["hit_id"].map(lambda x: translation_dict[x])
        d_motif_per_scan = {}
        for i in screening_hits["scan"].unique():
            d_motif_per_scan[i] = tuple(screening_hits["ref_motif_id"][screening_hits["scan"] == i])
        dataframe['temp_indexcopy'] = dataframe.index
        dataframe["MS2LDA:allmotifs"] = dataframe['temp_indexcopy'].map(lambda x: map_motifs(x,d_motif_per_scan))
        dataframe = dataframe.drop('temp_indexcopy',axis=1)
    return dataframe
# integrate_df_values_if_shared_other_value(integrated_df,plastchem_df,canonical_smiles,csifingerid:smiles,plastchem:hazard_score,hazard_score)
# def integrate_df_values_if_shared_other_value(df1,df2,col_sh1,colsh2,newcol,colint2):
# #    if sh1 = sh2, col1 = col2

def create_network_from_mgf(mgf: str) -> nx.Graph:
    """process spectra from MGF, calculate similarity and create network from those scores"""
    spectrums: list[matchms.Spectrum] = load_from_mgf(mgf)
    spectrums = [peak_processing(s) for s in spectrums] 
    similarity_measure = ModifiedCosine(tolerance=0.02)
    scores: matchms.Scores = calculate_scores(spectrums, spectrums, similarity_measure,is_symmetric=True)
    scores.to_pickle(f'{mgf[:-4]}_scores.pickle')
    # scores.filter_by_range(name="ModifiedCosine_matches",low=6,below_operator="<")

    network: nx.Graph = create_network_from_scores(scores)
    return network

def peak_processing(spectrum: matchms.Spectrum) -> matchms.Spectrum:
    """Apply filters to spectrum"""
    spectrum = remove_peaks_around_precursor_mz(spectrum,mz_tolerance=17)
    spectrum = remove_peaks_outside_top_k(spectrum,k=6,mz_window=50)
    return spectrum

def create_network_from_scores(scores: matchms.Scores ) -> nx.Graph:
    """Filters scores and creates network"""
    #scores: matchms.Scores = scores_from_pickle("/lustre/BIF/nobackup/hendr218/temp/scores_001")
    newscores = scores
    newscores.filter_by_range(name="ModifiedCosine_matches",low=6,below_operator="<")
    if "scans" in newscores.queries[0].metadata:
        ms_network = SimilarityNetwork(identifier_key="scans",score_cutoff=0.7,top_n=10,link_method="mutual")
    elif "feature_id" in newscores.queries[0].metadata:
         ms_network = SimilarityNetwork(identifier_key="feature_id",score_cutoff=0.7,top_n=10,link_method="mutual")
    else:
        raise ValueError("'Scans' or 'feature_id' required in mgf to identify features in network by other tools")
    ms_network.create_network(newscores,score_name="ModifiedCosine_score")
    return ms_network


def parse_cramer_classifications(fn_cramer_csv: str) -> pd.DataFrame:
    """parses cramer classification and smiles per compound from input csv"""
    with open(fn_cramer_csv) as file:
        l_cramer = []
        for line in file:
            #skip the first line
            if line.startswith('CRAMERFLAGS,Cramer rules'):
                continue
            cramer_class,x,cramer_smiles = line.split(',')[1:4]
            l_cramer.append((cramer_class,cramer_smiles))
    df_cramer = pd.DataFrame(l_cramer,columns=("cramer_classification","SMILES"))
    return df_cramer

def get_merging_settings(config: dict,target_integration_col: str) -> tuple[str,str,dict]:
        """Gets columns to integrate on and column translations from config"""
        translations: dict[Any] = config.get("translations", {})
        merging_transl: dict[Any]  = translations.get("merging", {})
        column_transl: dict[Any] | str = translations.get("columns", "N/A")
        source_integration_col: Any | str = merging_transl.get(target_integration_col,"N/A")
        return target_integration_col,source_integration_col,column_transl
#target_int_col,source_int_col,col_translator

def map_motifs(x: Any, motifs_per_scan: dict[str,list]) -> dict | str:
    """Helper to map finding motifs per scan for MSLDA to dataframe"""
    if x in motifs_per_scan:
        return motifs_per_scan[x]
    else:
        return "N/A"

def parse_classyfire_sdf(sdf_fn: str) -> pd.DataFrame:
    """Converts SDF from classyfire to dataframe containg classifications and inchis"""
    cf_columns = ("inchi","smiles","cf_kingdom","cf_superclass","cf_class","cf_subclass")
    df_sdf = rdkit.Chem.PandasTools.LoadSDF(sdf_fn)
    df_sdf['new'] = df_sdf['InChIKey'].map(parse_cf_string)
    parsed_df = pd.DataFrame(df_sdf['new'].tolist(),columns=(cf_columns))
    return parsed_df
    
def parse_cf_string(cf_entry: str):
    """Takes classyfire SDF output entry and parses classifications and inchi"""
    split_entry = cf_entry.split('> <Intermed')[0].split('\n')
    match len(split_entry):
        case 12:
            inchi,a,smiles,b, cf_kingdom, c, cf_superclass,d,cf_class,e,cf_subclass,f = split_entry
        case 10:
            inchi,a,smiles,b, cf_kingdom, c, cf_superclass,d,cf_class,f = split_entry
            cf_subclass = "N/A"
        case 8:
            inchi,a,smiles,b, cf_kingdom, c, cf_superclass,f = split_entry
            cf_subclass = cf_class = "N/A"
        case 6:
            inchi,a,smiles,b, cf_kingdom,f = split_entry
            cf_subclass = cf_class = cf_superclass = "N/A"
        case _:
            inchi = cf_kingdom = cf_subclass = cf_class = cf_superclass = "N/A"
    return (inchi,smiles,cf_kingdom,cf_superclass,cf_class,cf_subclass)   
# plastchem_searcher = DfSearcher(datab=plastchem_db) 
# col_query = integrated_df.loc[:,["csifingerid:smiles"]]
# for new_name,plastchem_name in [
#             ("plastchem:hazard_score","Hazard_score"),
#             ("plastchem:PlastChem_lists","PlastChem_lists"),
#             ("plastchem:use","Use"),
#             ("plastchem:pcname","pubchem_name")]:
    
#     integrated_df[new_name] = col_query.map(lambda x: p.process(x,prop=plastchem_name))



def temp_metadata_adder(dataframe: pd.DataFrame,input_mgf: str,metadata_csv: str,quant_table: str) -> nx.Graph:
    """takes input mgf,df and adds metadata"""
    with open(input_mgf) as file:
        A = file.read()
        entries = A.split("BEGIN IONS")[1:]

    id_files_dict = {}
    # rt_dict = {}
    # mass_dict = {}
    for entry in entries:
        feature_id = entry.split('\n')[1][11:] 
        if "FILENAME" in entry:
            files = entry.split("FILENAME")[1][1:].split('\n')[0]
            id_files_dict[feature_id] = files.split(";")

    metadata_dict = {}
    with open(metadata_csv) as file:
        for line in file.readlines()[1:]:
            name,plastic_type,plastic_state = line.split(',')
            plastic_state=plastic_state[:-1]
            polarity = "pos" if "pos" in name else "neg"
            if plastic_state == "Blank":
                state_type = "blank"
            elif plastic_state == "procBlank":
                state_type = "proc_blank"
            else:
                state_type = f"{plastic_type}_{plastic_state}"
            if not name in metadata_dict:
                metadata_dict[name] = state_type      

    id_statetype = {}
    for id, files in id_files_dict.items():
        id_statetype[id] = ""
        for file in files: 
            statetype = metadata_dict[file]
            if not statetype in id_statetype[id]:
                id_statetype[id] += f"{statetype}, "

    #  for i in id_statetype["id"].unique():
    
    quant_table =  pd.read_csv(quant_table,low_memory=False)
    samples = [key for key in quant_table.keys() if ("mzML:area" in key) & (not "Blank" in key)]
    samples.append('id')
    blanks = [key for key in quant_table.keys() if ("mzML:area" in key) & ("Blank" in key)]
    blanks.append('id')
    s_areas = quant_table.loc[:,samples].set_index('id').transpose()
    b_areas = quant_table.loc[:,blanks].set_index('id').transpose()

    s_area_tresholded = s_areas
    is_blank_dict: dict[str,bool] = {}
    for id in s_areas.keys():
        mean = s_areas[id].mean()
        s_area_tresholded[id] = s_areas[id].map(lambda x: True if mean < 0.3 * x else False)
    for id in s_areas.keys():
        mean = quant_table.loc[:,samples].set_index('id').transpose()[id].fillna(0).mean()
        blank_max = quant_table.loc[:,blanks].set_index('id').transpose()[id].fillna(0).max()
        is_blank: bool = mean<10*blank_max
        is_blank_dict[id] = is_blank
    dataframe["is_blank"] = dataframe.index.map(lambda x: is_blank_dict[int(x)]) 
    
    # feature_metadata_dict = {}
    # for id, files in s_area_tresholded.to_dict().items():
    #     id = str(id)
    #     feature_metadata_dict[id] = ""
    #     for file,val in files.items(): 
    #         if (val) & (not file in feature_metadata_dict[id]):
    #             feature_metadata_dict[id] += f"{metadata_dict[file[9:-5]]}, "
    feature_metadata_dict = {}
    for id, files in s_area_tresholded.to_dict().items():
        id = str(id)
        feature_metadata_dict[id] = []
        for file,val in files.items(): 
            metadata: str = metadata_dict[file[9:-5]]
            if (val) & (metadata not in feature_metadata_dict[id]):
                feature_metadata_dict[id].append(metadata)

    dataframe["all_metadata"] = dataframe.index.map(lambda x: feature_metadata_dict.get(x,"N/A"))
    for metadata_type in ["PE_PA_Dec", "PE_PA_PC","PE_Dec", "PE_PC", "PE_PET_PC", "PE_PET_Dec"]:
        dataframe[metadata_type] =  dataframe["all_metadata"].map(lambda x: metadata_type in x)
    for plastic_type in ["PE","PE_PA","PE_PET"]:
        dataframe[plastic_type] = dataframe["all_metadata"].map(lambda all_metadata: True if plastic_type in ["_".join(metadata_item.split("_")[0:-1]) for metadata_item in all_metadata] else False)
    for colname,plastic_state in [("Decontaminated","Dec"),("Postconsumer","PC")]:
        dataframe[colname] = dataframe["all_metadata"].map(lambda all_metadata: True if plastic_state in [metadata_item.split("_")[-1] for metadata_item in all_metadata] else False)
    dataframe["Introduced_by_dec"] = dataframe.apply(lambda row: introduced_by_decontamination_mapper(row),axis=1)
    dataframe["Removed_by_dec"] = dataframe.apply(lambda row: removed_by_decontamination_mapper(row),axis=1)
    dataframe["Kept_by_dec"] = dataframe.apply(lambda row: kept_by_decontamination_mapper(row),axis=1)

    # dataframe["PE"] = dataframe["all_metadata"].map(lambda x: True if ("PE_Dec" in x or "PE_PC" in x) else False)
    # dataframe["plastic_types"] = dataframe["all_metadata"].apply(plastic_type_mapper)
    dataframe["plastic_types"] = dataframe.apply(plastic_type_mapper,axis=1)
    return dataframe

def introduced_by_decontamination_mapper(row)->bool:
    metadata_items = [metadata_item.split("_")[-1] for metadata_item in row["all_metadata"]]
    if "Dec" in metadata_items and not "PC" in metadata_items: 
        return True
    else:
        return False
    
def removed_by_decontamination_mapper(row)->bool:
    metadata_items = [metadata_item.split("_")[-1] for metadata_item in row["all_metadata"]]
    if "PC" in metadata_items and not "Dec" in metadata_items: 
        return True
    else:
        return False
    
def kept_by_decontamination_mapper(row)->bool:
    metadata_items = [metadata_item.split("_")[-1] for metadata_item in row["all_metadata"]]
    if "PC" in metadata_items and "Dec" in metadata_items: 
        return True
    else:
        return False

def plastic_type_mapper(row) -> str:
    if row["PE_PA"] & (row["PE_PET"]):
        return "PE-PET-PA" if row["PE"] else "PET-PA"
    if row["PE_PA"]:
        return "PE-PA" if row["PE"] else "PA"
    if row["PE_PET"]:
        return "PE-PET" if  row["PE"] else "PET"
    if row["PE"]:
        return "PE"
    else:
        return "N/A"
# def plastic_type_mapper(row) -> str:
#     if row(["PE_PA"]) & row["PE_PET"]:
#     if ("PE_PA" in all_types) & ("PE_PET" in all_types):
#         return "PE-PET-PA" if ("PE" in all_types) else "PET-PA"
#     if ("PE_PA" in all_types):
#         return "PE-PA" if ("PE" in all_types) else "PA"
#     if ("PE_PET" in all_types):
#         return "PE-PET" if  ("PE" in all_types) else "PET"
#     if "PE" in all_types:
#         return "PE"
#     else:
#         return "N/A"



    # p = DfSearcher(datab=plastchem_db) 
    # col_query = integrated_df.loc[:,["csifingerid:smiles"]]
    # for new_name,plastchem_name in [
    #         ("plastchem:hazard_score","Hazard_score"),
    #         ("plastchem:PlastChem_lists","PlastChem_lists"),
    #         ("plastchem:use","Use"),
    #         ("plastchem:pcname","pubchem_name")]:
    #     integrated_df[new_name] = col_query.map(lambda x: p.process(x,prop=plastchem_name))


'''Script to analyze NIAS from plastic data

First provide a settings file and a dataset. 
Dataset files should be in mzML format if MZMine is used, otherwise MGFs are required.
Generate a workflow instance (a combination of dataset and settings) with the workflowrunner class.
Then run those instances with workflowrunner.run()
Repeat for all settings and datasets as required

'''

import json
import pickle
from time import ctime
from typing import Any,Iterable
import os.path
import sys
import xml.etree.ElementTree as ET
from pathlib import Path

failed_imports = []
try:
    import MS2LDA
except ImportError:
    failed_imports.append("MS2LDA")

from .integrating import *
from .parsing import *
# from src.myworkflow.visualisation import *
# from src.myworkflow.unrefined import *
from .running import * #run_mzmine, run_sirius, run_mzmine run_toxtree,run_classyfire, run_ms2lda_post_to_classyfire, check_classyfire_done, get_classyfire_results
from .util import validate_dictkeys,json_from_string_or_file,get_filelist_from_folder,convert_missing,integrate_df_cols_to_df
from .external import batched,filter_component,prune_component,get_edges_of_component
from .integrating import *
from .parsing import *
# from src.myworkflow.visualisation import *
# from src.myworkflow.unrefined import *
# from src.myworkflow.running import * #run_mzmine, run_sirius, run_mzmine run_toxtree,run_classyfire, run_ms2lda_post_to_classyfire, check_classyfire_done, get_classyfire_results

class MetaboMix():
    '''Holds the full workflow

    Used to  use the workflow for a specific combination of parameters and settings. 

    Parameters
    ----------
    settings : str | dict
        fn of settings file or dict in json format
        should contain settings defined in workflowsettings required settings class property     

    Attributes
    ----------
    available_tools : list
        Defines which tools can be run. runtool should also be set True in settings file to actually run the tool.
    available_integrations : list
        Defines which integrations can be selected. Integrate_tool should also be set True in settings file to actually integrate the tool.

    '''
    
    available_tools: list[str] = ["mzmine","ms2lda","sirius","toxtree","classyfire", "matchms"]
    available_integrations: list[str] = ["mzmine","ms2lda","sirius","toxtree","classyfire", "matchms","plastchemdb","sirius_db"]

    def __init__(self,settings_json: str | dict) -> None:
        self.settings: RecipeHolder = RecipeHolder(settings_json,available_tools=MetaboMix.available_tools,available_integrations=MetaboMix.available_integrations)
        self.name = self.settings.name
        self.output_folder = self.settings.output_folder
        self.graph: nx.Graph = nx.Graph
        self.network_df, self.network_edgelist = pd.DataFrame,nx.edgelist

    def get_internal_settings(self,settings_json: str | dict) -> dict:
        """Check if internal settings were provided and return them as a json dict
        
        Parameters
        ----------
        settings_json : str | dict
            fn of settings file or settings dict in json format

        Returns
        -------
        Settings dict in json format
        """
        try:
            fn = json_from_string_or_file(settings_json)["paths"].get("internal_settings")
        except KeyError:
            print("settings file must contain 'internal_setting' path in paths")
        return json.load(fn)
    
    def run_all(self) -> None:
        self.run_and_integrate_per_depth()
        print("Sucesfully ran and integrated all tools")

    def get_depths(self,tools: Iterable,goal:str) -> dict[Any,str]:
        """"""
        depths: dict = {}
        for tool in tools:
            config = self.settings.config.get(tool,{})
            goal_depth = config.get(goal,{}).get("depth")
            depth = goal_depth if goal_depth else config.get("depth","N/A")
            match depth:
                case  "1" | 1:
                    depths[tool] = 1
                case  "2" | 2:
                    depths[tool] = 2
                case  "3" | 3:
                    depths[tool] = 3
                case  "4" | 4:
                    depths[tool] = 4
                case  "N/A":
                    depths[tool] = "N/A"
        return depths

    def run_and_integrate_per_depth(self) -> None:
        """For every depth in ascending order, run and integrate tools that are of that depth, finally produce the integrated network graphml file.

        Every tool should have a depth for running and integration specified in internal settings. For every depth, first the tools will be ran, then integrated. 
        This is only attempted for tools that were selected in class property `available tools` or `available integrations`. 
        As the first thing on depth 3, the graph is made so any tools requiring the graph should be of depth 3+.
        As the first thing on depth 4, the network is made so any tools requiring the network should be of depth 4+.

        """

        running_depths: dict[Any,str] = self.get_depths(self.settings.to_run,goal="running")
        integration_depths: dict[Any,str] = self.get_depths(self.settings.to_integrate,goal="integration")
        for current_depth in [1,2,3,4,"N/A"]:
            running: set[Any] = {tool for tool,depth in running_depths.items() if depth==current_depth}
            integrating: set[Any] = {tool for tool,depth in integration_depths.items() if depth==current_depth}               
            if current_depth == 3:
                self.graph = nx.read_graphml(self.settings.paths.get("base_network"))
            if current_depth == 4:   
                self.network_df, self.network_edgelist = network_to_edgelist_and_nodes_df(self.graph)
                self.integration_col = "smiles"  
                self.network_df["smiles"] = self.network_df.apply(self.get_consensus_smiles,axis=1)
                self.network_df["Molecular formula"] = self.network_df.apply(self.get_consensus_formula,axis=1)
            for tool in running:
                self.run_tool(tool)
            for integration in integrating:
                self.integrate_tool(integration)

        for cf_class in ["subclass","class","superclass"]:
                    self.network_df[cf_class] = self.network_df.apply(lambda x: self.get_consensus_class(row=x,classtype=cf_class),axis=1)
       # target_df[target_df["Molecular formula"].notna()].merge(source_df[source_df["molecular_formula"].notna()],left_on="Molecular formula",right_on="molecular_formula")
        self.produce_integrated_graphml()
        
    def get_consensus_formula(self,row):
        """Get formula from database sirius formula if exists, otherwise normal formula"""
        # if row["library_smiles"]
        if not row.get("sirius:molecularFormula"): 
            row["sirius:molecularFormula"] = "N/A"
        if row.get("sirius_db:molecularFormula"):
            return row["sirius_db:molecularFormula"] if isinstance(row["sirius_db:molecularFormula"],str) else row["sirius:molecularFormula"]
        else:
            return row["sirius:molecularFormula"] 
        
    def get_consensus_smiles(self,row):
        """Get smiles from database sirius smiles if exists, otherwise normal smiles"""
        # if row["library_smiles"]
        if not row.get("csifingerid:smiles"): 
            row["csifingerid:smiles"] = "N/A"
        if row.get("csifingerid_db:smiles"):
            return row["csifingerid_db:smiles"] if isinstance(row["csifingerid_db:smiles"],str) else row["csifingerid:smiles"]
        else:
            return row["csifingerid:smiles"] 
    
    def get_consensus_class(self,row,classtype):
        if not row.get(f"canopus:CF_{classtype}"): 
            row[f"canopus:CF_{classtype}"] = "N/A"
        if row.get(f"cf:{classtype}"):
            return row[f"cf:{classtype}"] if isinstance(row[f"cf:{classtype}"],str) else row[f"canopus:CF_{classtype}"]
        else: 
            return row[f"canopus:CF_{classtype}"]
    
    def get_features_per_motif(self,motif_col: pd.DataFrame) -> dict[str,list[str]]:
        features_per_motif= {}
        for feature_id, motifs in motif_col.to_dict().items():
            for motif in motifs:
                if not motif in features_per_motif:
                    features_per_motif[motif] = []
                    features_per_motif[motif].append(feature_id)
                if not feature_id in features_per_motif[motif]:
                    features_per_motif[motif].append(feature_id)
        return features_per_motif

    def produce_integrated_graphml(self) -> None:
        """Cleans, filters and creates graphml from network DF and edges
        
        Topology filter is always used: cluster can be only of 100 nodes. 
        Converts all missing values to "N/A" and turns all properties into strings for easier reading in cytoscape!
        """
        integrated_df: pd.DataFrame = self.network_df
        integrated_network = nx.from_pandas_edgelist(self.network_edgelist,edge_attr=True)
        integrated_df_as_strings = integrated_df.astype(str)
        integrated_df_cleaned = integrated_df_as_strings.map(convert_missing)
        nodes_dict = integrated_df_cleaned.to_dict(orient='index')
        for node_ID in nodes_dict:
            if not node_ID in integrated_network:
                integrated_network.add_node(node_ID)
            #nodes_dict[node_ID] = {}
            for attribute in nodes_dict[node_ID]:
                integrated_network.nodes[node_ID][attribute] = nodes_dict[node_ID][attribute]
        filter_component(integrated_network,100)

        nx.write_graphml(integrated_network, f"{self.output_folder}/{self.name}_premotifs.graphml")

        if "MS2LDA" in self.settings.to_integrate:
            df_motifs = integrated_df["MS2LDA:allmotifs"][integrated_df["MS2LDA:allmotifs"] != "N/A"]
            features_per_motif = self.get_features_per_motif(df_motifs)
            for motif in features_per_motif:
                combinations = list(itertools.combinations(features_per_motif[motif],2))
                # print(combinations)
                combinations = [tuple([combination[0],combination[1],{motif:True}]) for combination in combinations]
                integrated_network.add_edges_from(combinations,motif=motif,type="motif")
            nx.write_graphml(integrated_network, f"{self.output_folder}/{self.name}_postmotifs.graphml")

    def run_tool(self,tool: Any) -> None:
        """Get running parameters and run selected tool
        
        Here, all running behavirous for a specific tool is defined. First the required parameters are obtained from properties that were set & checked in `WorkflowSettings`.
        Simple format converions of input are done here as well.  
        No order is defined here: use depth in internal settings for that.
        Will only do something for tools defined in `available_tools` class property.

        """
        match tool:
            case "classyfire":
                #print('ran classyfire (not, debug)')
                input_csv = self.create_smiles_csv_from_df(output_name = f"{self.output_folder}/classyfire_input.csv",header=False)
                output_path = self.settings.paths.get("classyfire_output")
                print('running classyfire')
                run_classyfire(input_path=input_csv,output_path=output_path)
                # parse_classyfire_sdf("/lustre/BIF/nobackup/hendr218/temp/sdftest.txt")
            case "matchms":
                print("running matchms")
                try:
                    graph: nx.Graph = create_network_from_mgf(self.settings.paths.get("input_mgf"))
                except ValueError:
                    raise InputError("Input MGF was not valid")
                output_path = self.settings.paths.get("base_network")
                graph.export_to_graphml(output_path)
                #nx.write_graphml(graph, output_path)
            case "ms2lda":
                output_path: str = self.settings.paths.get("ms2lda_output")
                self.settings.ms2lda["dataset_parameters"]["output_folder"] = output_path
                ms2lda_params: dict = self.settings.get_ms2lda_params()
                input_mgf: str = self.settings.paths.get("input_mgf")
                #print("ran ms2lda (not, debug)")
                print('running ms2lda')
                run_ms2lda(dataset=input_mgf,params=ms2lda_params)
            case "mzmine":
                mzmine_location = self.settings.paths.get("mzmine_location")
                mzmine_userfile_location = self.settings.paths.get("mzmine_userfile_location")
                base_batchfile = self.settings.paths.get("mzmine_base_batchfile")
                output_loc = f'{self.output_folder}/mzmine'
                #get filelist if it was provided, else fall back to the folder
                self.data: list = self.settings.paths.get("file_list", get_filelist_from_folder(self.settings.paths.get("data_folder")))
                batchfile_writer = mzbatch_writer(base_mzbatch=base_batchfile,data=self.data,mzmine_params=self.settings.mzmine)
                batchfile = batchfile_writer.write_mzbatch(output_filename=f'{output_loc}_mzbatch.mzbatch')
                #print("ran mzmine, not, debug")
                print('running mzmine')
                run_mzmine(batchfn=batchfile,mzmine_output_loc=output_loc,mzmine_location=mzmine_location,mzmine_userfile_location=mzmine_userfile_location,temp_folder=f'{output_loc}/temp/')
            case "sirius":
                output_loc: str = self.settings.paths.get("sirius_output")
                sirius_path: str = self.settings.paths.get("sirius_path")
                input_mgf: str= self.settings.paths.get("input_mgf")
                instrument: str = self.settings.sirius.get("instrument")
                formula_db: str = self.settings.sirius.get("formula_db")
                #print("ran sirius (not, debug)")
                print('running sirius')
                run_sirius(input_path=input_mgf,output_path=output_loc,sirius_path=sirius_path,formula_db=formula_db,instrument=instrument)
            case "toxtree":
                toxtree_loc = self.settings.paths.get("toxtree_path")
                toxtree_module = self.settings.toxtree.get("module")
                config: dict = self.settings.config.get("toxtree",{}).get("running",{}).get("translations",{})
                if not toxtree_module in config.get("modules",{}):
                    raise Settingserror(f"toxtree module {toxtree_module} not supported, add to module translation in internal settings to support")
                toxtree_module = config.get("modules",{}).get(toxtree_module)
                input_csv = self.create_smiles_csv_from_df(output_name = f"{self.output_folder}/toxtree_input.csv",header=True)
                output_path = f"{self.output_folder}/toxtree_results.csv"
                #print("ran toxtree (not, debug)")
                print('running toxtree')
                run_toxtree(input_path=input_csv,output_path=output_path,toxtree_path=toxtree_loc,module_path=toxtree_module)
       
    def integrate_tool(self,tool: Any) -> None:
        """Get needed paramaters and integrate selected tool
        
        Here, all integration behaviour for a specific tool is defined. First the required parameters are obtained from properties that were set & checked in `WorkflowSettings`.
        Simple format converions of input are done here as well.  
        No order is defined here: use depth in internal settings for that.
        Will only do something for tools defined in `available_integrations` class property.

        """
        match tool:
            case "classyfire":
                print('integrating classyfire')
                output_folder: str = self.settings.paths.get("classyfire_output")
                classyfire_records: pd.DataFrame= parse_classyfire_sdf(output_folder)
                config: dict  = self.settings.config.get("classyfire", {}).get("integration", {})
                integration_settings: tuple[dict,str] = get_merging_settings(config,self.integration_col)
                self.network_df: pd.DataFrame = integrate_df_cols_to_df(self.network_df,classyfire_records, *integration_settings)
            case "ms2lda":
                print('integrating ms2lda')
                output_folder = self.settings.paths.get("ms2lda_output")
                dataset = self.settings.paths.get("input_mgf")
                output_database = self.settings.paths.get("ms2lda_results_db")
                self.network_df = integrate_ms2lda_to_df(self.network_df,dataset,output_folder,output_database)
            case "mzmine":
                #temp to get some quick results, will be improved later dataframe: pd.DataFrame,input_mgf: str,metadata_csv: str,quant_table: str
                print('integrating mzmine')
                input_mgf = self.settings.paths.get("sirius_mgf")
                quant_table = self.settings.paths.get("quant_table")
                metadata_csv = self.settings.paths.get("metadata_csv")
                self.network_df = temp_metadata_adder(self.network_df,input_mgf,metadata_csv,quant_table)
            case "sirius":
                print('integrating sirius')
                config: dict = self.settings.config.get("sirius",{}).get("integration",{}).get("translations",{})
                translators = config.get("sirius",{}),config.get("csi:fingerid",{}),config.get("canopus",{})
                outputs = self.settings.paths.get("sirius_tool_output"), self.settings.paths.get("csi:fingerid_output"), self.settings.paths.get("canopus_output")
                self.graph = integrate_sirius_to_graph(graph=self.graph,translators=translators,outputs=outputs)
            case "sirius_db":
                print('integrating sirius from db')
                config: dict = self.settings.config.get("sirius_db",{}).get("integration",{}).get("translations",{})
                translators = config.get("sirius",{}),config.get("csi:fingerid",{}),config.get("canopus",{})
                outputs = self.settings.paths.get("sirius_tool_db_output"), self.settings.paths.get("csi:fingerid_db_output"), self.settings.paths.get("canopus_db_output")
                self.graph = integrate_sirius_to_graph(graph=self.graph,translators=translators,outputs=outputs)
            case "toxtree":
                print('integrating toxtree')
                toxtree_output: str = self.settings.paths.get("toxtree_output")
                toxtree_df: pd.DataFrame = parse_cramer_classifications(toxtree_output)
                config: dict  = self.settings.config.get("toxtree", {}).get("integration", {})
                integration_settings = get_merging_settings(config,self.integration_col)
                self.network_df = integrate_df_cols_to_df(self.network_df,toxtree_df, *integration_settings)
            case "plastchemdb":
                print('integrating plastchem')
                config: dict  = self.settings.config.get("plastchemdb", {}).get("integration", {})
                plastchem_db: str = self.settings.paths.get("plastchem_path")
                plastchem_df =  pd.read_csv(plastchem_db, sep='\t', encoding='windows-1251', low_memory=False, header=1)
                integration_settings = get_merging_settings(config,self.integration_col) 
                self.network_df = integrate_df_cols_to_df(self.network_df,plastchem_df, *integration_settings)

    def create_smiles_csv_from_df(self,header: bool,output_name: str) -> bool:
        """Creates a smiles file from a dataframe column with rows on newlines"""
        network_df = self.network_df
        #df_smiles = network_df[network_df['csifingerid:smiles'].notna()].rename(columns={"csifingerid:smiles": "smiles"})      
        df_smiles = network_df[network_df['smiles'].notna()]  
        df_smiles.to_csv(output_name,columns=["smiles"],header=header, index=False,na_rep="N/A")
        return output_name 
    
class Settingserror(Exception):
    """Error for indicating invalid settings files"""
    pass

class InputError(Exception):
    """Error for indicating invalid input files"""
    pass

class RecipeHolder:
    """Class to process and hold settings for the workflow

    First check if the always required fields from `required_settings' and `required_paths` class attributes are present. 
    If so, loads paths and adds new ones depending on the selected tools. (i.e. if a tool creates a file that might be needed downstream, it should add it to paths)
    Then, checks if all tools have the input paths that they need as defined in the internal_settings file
    Finally, tests if 

    Parameters
    ----------
    settings : str
        fn of settings file
    available_tools : list[str]
        tools supported by worklflow: others will be ignored
    available_integrations : list[str]
        integrations supported by worklflow: others will be ignored

    Attributes
    ----------
    required_settings : list[str]
        settings that must be provided in settings json 
    required_settings tools : list[str]
        paths that must be provided in settings settings json paths field

    """
    required_settings: list[str] = ["paths","run_tools","integrate_tools"]
    required_paths: list[str] = ["base_output_folder","internal_settings"]

    def __init__(self,settings_json:str | dict,available_tools: list[str],available_integrations: list[str]) -> None:
        self.input: dict = json_from_string_or_file(settings_json)   
        self.validate_required_settings(settings=self.input)
        self.config: dict = json_from_string_or_file(self.input["paths"]["internal_settings"])

        #Add properties for convenience in use and code
        self.paths: dict = self.input["paths"]
        self.name = self.paths.get("name",ctime().replace(" ","_"))
        self.output_folder = f'{self.paths.get("base_output_folder")}/{self.name}'
        Path(self.output_folder).mkdir(exist_ok=True)
        self.to_run: list[str] = self.select_used_tools(available_tools=available_tools,setting="run_tools")
        self.to_integrate: list[str] = self.select_used_tools(available_tools=available_integrations,setting="integrate_tools")

        #Let tools set additional settings and then check if all requirements are met
        self.set_tool_settings(set(self.to_run+self.to_integrate))
        self.check_tool_requirements(tools=self.to_run,goal="running")
        self.check_tool_requirements(tools=self.to_integrate,goal="integration")

        self.check_tool_requirements(tools=["all_tools"],goal="running")
        self.check_tool_requirements(tools=["all_tools"],goal="integration")


    def validate_required_settings(self,settings: dict) -> None:
        """Check if settings json contains required fields"""
        if not validate_dictkeys(settings,RecipeHolder.required_settings):
            raise Settingserror(f"settings json must contain {', '.join(RecipeHolder.required_settings)}")
        if not validate_dictkeys(settings["paths"],RecipeHolder.required_paths):
            raise Settingserror(f"settings json paths must contain {', '.join(RecipeHolder.required_paths)}")
    

    def check_tool_requirements(self,tools: Iterable[str],goal: str) -> None:
        for tool in tools:
            requirements = self.get_tool_requirements(tool,goal)
            self.validate_tool_requirements(tool,requirements)               
    
    def set_tool_settings(self,tools: Iterable[str]) -> None:
        """set tool settings as property and add any settings that depend on the tool""" 
        for tool in tools:
            tool_settings = self.input.get(tool,{})
            match tool:
                case "mzmine":
                    self.mzmine = tool_settings
                    self.paths["input_mgf"] = f'{self.output_folder}/mzmine/mzmine_iimn_fbmn.mgf'
                    self.paths["sirius_mgf"] = f'{self.output_folder}/mzmine/mzmine_sirius.mgf'
                    self.paths["quant_table"] = f'{self.output_folder}/mzmine/mzmine_iimn_fbmn_quant_full.csv'
                case "sirius":
                    self.sirius = tool_settings
                    sirius_loc = f'{self.output_folder}/sirius/'

                    Path(sirius_loc).mkdir(exist_ok=True)
                    
                    if "sirius" in self.to_run:
                        self.paths["csi:fingerid_output"] = f"{sirius_loc}structure_identifications.tsv"
                        self.paths["sirius_tool_output"] = f"{sirius_loc}formula_identifications.tsv"
                        self.paths["canopus_output"] = f"{sirius_loc}canopus_structure_summary.tsv"

                    self.paths["sirius_output"] = sirius_loc
                case "sirius_db":
                    self.siriusdb = tool_settings
                    siriusdb_loc = f'{self.output_folder}/sirius_db/'
                    self.paths["canopus_db_output"] = f"{siriusdb_loc}canopus_structure_summary.tsv"
                    self.paths["csi:fingerid_db_output"] = f"{siriusdb_loc}structure_identifications.tsv"
                    self.paths["sirius_tool_db_output"] = f"{siriusdb_loc}formula_identifications.tsv"
                    self.paths["sirius_db_output"] = siriusdb_loc
                case "matchms":
                    self.matchms = tool_settings
                    self.paths["base_network"] = f'{self.output_folder}/base_network.graphml'
                case "toxtree":
                    self.toxtree = tool_settings
                    self.paths["toxtree_output"] = f"{self.output_folder}/toxtree_results.csv"
                    print("warning: classyfire requires SMILES to be in the dataframe, currently checks for that are not implemented")
                case "classyfire":
                    self.classyfire = tool_settings
                    self.paths["classyfire_output"] = f"{self.output_folder}/classyfire_results.sdf"
                    print("warning: classyfire requires SMILES to be in the dataframe, currently checks for that are not implemented")
                case "ms2lda":
                    self.ms2lda = tool_settings
                    ms2lda_loc = f"{self.output_folder}/ms2lda"
                    self.paths["ms2lda_output"] = f"{ms2lda_loc}"
                    self.paths["ms2lda_results_db"] = f"{self.output_folder}/ms2lda/motifDB_optimized.xlsx"

    def get_tool_requirements(self,tool: str, goal:str) -> tuple[list,list,list,list]:
        """Get requirements if available, otherwise return emoty list"""
        config = self.config.get(tool,{}).get(goal,{}).get("requirements",{})
        required_paths: Iterable[Any] = config.get("paths",[])
        required_settings: Iterable[Any] = config.get("settings",[])
        required_optional_paths: Iterable[Iterable[Any]] = config.get("optional_paths",[])
        required_optional_settings: Iterable[Iterable[Any]] = config.get("optional_settings",[])
        required_modules: Iterable[Iterable[Any]] = config.get("modules",[])
        return required_paths,required_settings,required_optional_paths,required_optional_settings,required_modules
    
    def validate_tool_requirements(self,tool, requirements):
        """Raise errors if required settings are not present in settings"""
        required_paths,required_settings,required_optional_paths,required_optional_settings,required_modules = requirements
            # if not validate_dictkeys(self.input["paths"],required_paths):
        if not validate_dictkeys(self.paths,required_paths):
            raise Settingserror(f"all of {required_paths} required to run {tool}, but one is missing in paths settings")
        if not validate_dictkeys(self.input.get(tool),required_settings):
            raise Settingserror(f"One of {required_settings} missing in {tool} settings")
        for combination in required_optional_paths:
            if not validate_dictkeys(self.paths,combination):
                    raise Settingserror(f"Any one of {combination} required to run {tool}, but one is missing in paths settings")
        for combination in required_optional_settings:
            if not validate_dictkeys(self.input.get(tool),combination):
                    raise Settingserror(f"Any of {combination} is required in {tool} settings")
        for module in required_modules:
            if not module in sys.modules:
                raise Settingserror(f"all of {required_modules} required to run {tool}, but one is missing in imported modules")

    def select_used_tools(self,available_tools: list,setting: str) -> list:
        """Check which available tools were selected in settings file"""
        return [tool for tool in available_tools if self.input[setting].get(tool) == "True"]
    
    def get_ms2lda_params(self):
        """Get the parameters ms2lda needs froms settings"""
        preprocessing_parameters = self.ms2lda["preprocessing_parameters"]
        convergence_parameters = self.ms2lda["convergence_parameters"]
        annotation_parameters = self.ms2lda["annotation_parameters"]
        model_parameters = self.ms2lda["model_parameters"]
        train_parameters = self.ms2lda["train_parameters"]
        fingerprint_parameters = self.ms2lda["fingerprint_parameters"]
        dataset_parameters = self.ms2lda["dataset_parameters"]
        n_motifs = self.ms2lda["n_motifs"]
        n_iterations = self.ms2lda["n_iterations"]
        motif_parameter = self.ms2lda["motif_parameter"]

        return preprocessing_parameters,convergence_parameters,annotation_parameters,model_parameters,train_parameters,fingerprint_parameters,dataset_parameters,n_iterations,n_motifs,motif_parameter



import networkx as nx
from itertools import islice

def filter_component(G, max_component_size):
    """From GNPS! drop lowest edges from cluster if it exceeds max size """
    if max_component_size == 0:
        return

    big_components_present = True

    while big_components_present == True:
        big_components_present = False
        components = nx.connected_components(G)
        for component in components:
            if len(component) > max_component_size:
                prune_component(G, component)
                big_components_present = True
        
def prune_component(G, component, cosine_delta=0.02):
    component_edges = get_edges_of_component(G, component)

    min_score = 1000
    for edge in component_edges:
        min_score = min(min_score, edge[2]["weight"])

    cosine_threshold = cosine_delta + min_score
    for edge in component_edges:
        if edge[2]["weight"] < cosine_threshold:
            #print(edge)
            G.remove_edge(edge[0], edge[1])

def get_edges_of_component(G, component):
    component_edges = {}
    for node in component:
        node_edges = G.edges((node), data=True)
        for edge in node_edges:
            if edge[0] < edge[1]:
                key = edge[0] + "-" + edge[1]
            else:
                key = edge[1] + "-" + edge[0]
            component_edges[key] = edge

    component_edges = component_edges.values()
    return component_edges

#from 
def batched(iterable, n, *, strict=False):
    """from itertools python 3.12, split iterable in batches of size n"""
    # batched('ABCDEFG', 3) → ABC DEF G
    if n < 1:
        raise ValueError('n must be at least one')
    iterator = iter(iterable)
    while batch := tuple(islice(iterator, n)):
        if strict and len(batch) != n:
            raise ValueError('batched(): incomplete batch')
        yield batch




"""Module providings function for running various mass spectrometry tools

Included: mzmine, sirius, cramer, classyfire

"""

import json
from time import sleep
import requests
from typing import Any
import json
from time import ctime
from typing import Any
#from src.myworkflow.integrating import *
from .parsing import *
# from src.myworkflow.visualisation import *
from .util import validate_dictkeys,file_to_batches,run_subprocess,json_from_string_or_file,get_filelist_from_folder
from pathlib import Path
import xml.etree.ElementTree as ET
import os.path

import numpy as np
#import pymolnetenhancer as pme
import pandas as pd
from urllib.request import urlretrieve
import requests
import json
import networkx as nx
import subprocess
import rdkit
from time import sleep
import rdkit.Chem 
import rdkit.Chem.PandasTools
import matplotlib.pyplot as plt
import itertools
try:
    import MS2LDA
except ImportError:
    MS2LDA = None
import xml.etree.ElementTree as ET
from pathlib import Path

   
class mzbatch_writer():
        """Class to replace values from mzbatch with custom values"""
        searchterm_translator = {
        "min_samples":      (".//parameter[@name='Minimum aligned features (samples)']/abs",
                            ".//parameter[@name='Min samples filter']/parameter[@name='Min samples in all']/abs",),
        "feature_height":  (".//parameter[@name='Minimum absolute height']",),
        "RT_start":        (".//parameter[@name='Retention time']/min",),
        "RT_end":          (".//parameter[@name='Retention time']/max'",),
        "RTsearchrange":   (".//parameter[@name='Minimum search range RT/Mobility (absolute)']",),
        "top_edge_ratio":  (".//parameter[@name='Min ratio of peak top/edge']",),
        "chromtreshold":   (".//parameter[@name='Chromatographic threshold']",),
        }
        
        def __init__(self,base_mzbatch: str, data: list[str], mzmine_params: dict) -> None:
            self.base_mzbatch = base_mzbatch
            self.data = data
            self.params = mzmine_params


            tree = ET.ElementTree()
            tree.parse(self.base_mzbatch)
            self.tree = tree
            self.root: ET.Element = tree.getroot()

        def write_mzbatch(self,output_filename) -> str:
            """replace various parameters and write resulting batchfile"""
            self.replace_filenames()
            self.replace_params()
            
            # metadata=self.metadata
            # self.replace_metadata(root,metadata_fn)

            ET.indent(self.tree)
            self.tree.write(output_filename)

            return output_filename

        def replace_params(self) -> None:
            """look up mzbatch description for available params and use to set new values"""
            provided_params: list = self.get_provided_params(available_params=mzbatch_writer.searchterm_translator.keys())
            for provided_param in provided_params:
                searchterms = mzbatch_writer.searchterm_translator[provided_param] 
                value = self.params.get(provided_param)
                for searchterm in searchterms:
                    self.replace_param(param_name=searchterm,new_value=value)
               
  

        def get_provided_params(self,available_params) -> list:
            """Check which available mzmine params were selected in settings file"""
            provided_params: list = [param for param in available_params if self.params.get(param) == "True"]
            return provided_params

           
        
        def replace_param(self,param_name,new_value):
            """Replace element in mzbatch with new value"""
            for elem in self.do_mzbatch_search(param_name):
                elem.text = new_value

        def do_mzbatch_search(self,searcher_holder: list[str]) -> Any:
            "Find all instances of an element in mzbatch"
            all_findings: list[ET.Element] = []
            for searchterm in searcher_holder:
                for elem in self.root.findall(searchterm):
                    all_findings.append(elem)
                for elem in all_findings:
                    yield elem

        def replace_metadata(self,metadata_fn: str) -> None:
            """Replace metadata file with file from settings"""
            metadata_element: ET.Element | None = self.root.find(".//parameter[@name='Metadata file']/current_file") #root.findall("./batchstep/parameter/..[@name='File Names]"):
            if metadata_element is None:
                raise Settingserror("MZmine batchfile should contain 'Metadata file' parameter")
            metadata_element.text = metadata_fn   

        def replace_filenames(self) -> None:
            """Replace filenames from base mzbatch with custom ones"""
            oldfiles: list = self.root.findall(".//*[@name='File names']/") 
            fileloc: ET.Element | None = self.root.find(".//*[@name='File names']")
            if fileloc is None:
                raise Settingserror("MZmine batchfiles should contain 'File names' parameter")
            for oldfile in oldfiles:
                fileloc.remove(oldfile)
            for newfile in self.data:
                file_element = ET.SubElement(fileloc,"file")
                file_element.text=newfile



def run_mzmine(mzmine_location,mzmine_userfile_location,batchfn,temp_folder,mzmine_output_loc) -> bool:
    subprocess_command =  f"{mzmine_location} -user {mzmine_userfile_location} -b {batchfn} -o {mzmine_output_loc} -t {temp_folder}"
    run_subprocess(subprocess_command)
    return True
    
def run_sirius(input_path: str, output_path: str, sirius_path: str, **kwargs) -> bool:
    """"Run sirius from the shell

    Parameters
    ----------
    input_path
        file path to .MGF file, containing ..
    output path
        folder path to output folder
    instrument
        type of instrument, "orbitrap" or "qtof"

    Returns
    -------
    bool
        True if function sucesfully ran, False otherwise
    
    """
    #config = "--FormulaSearchSettings.applyFormulaConstraintsToBottomUp=false --IsotopeSettings.filter=true --UseHeuristic.useOnlyHeuristicAboveMz=650 --FormulaSearchDB=plastchem --Timeout.secondsPerTree=0 --FormulaSettings.enforced=H,C,N,O,F,P,I --Timeout.secondsPerInstance=0 --AlgorithmProfile=qtof --SpectralMatchingMassDeviation.allowedPeakDeviation=10.0ppm --AdductSettings.ignoreDetectedAdducts=false --AdductSettings.enforced=, --AdductSettings.prioritizeInputFileAdducts=true --UseHeuristic.useHeuristicAboveMz=300 --IsotopeMs2Settings=IGNORE --MS2MassDeviation.allowedMassDeviation=10.0ppm --SpectralMatchingMassDeviation.allowedPrecursorDeviation=10.0ppm --FormulaSearchSettings.performDeNovoBelowMz=0 --FormulaSearchSettings.applyFormulaConstraintsToDatabaseCandidates=false --EnforceElGordoFormula=true --NumberOfCandidatesPerIonization=1 --FormulaSettings.detectable=B,S,Cl,Se,Br --NumberOfCandidates=10 --AdductSettings.fallback=[[M+H]+,[M+Na]+,[M+K]+] --FormulaSearchSettings.performBottomUpAboveMz=Infinity --RecomputeResults=false spectra-search"
    instrument = kwargs["instrument"] if "instrument" in kwargs else "orbitrap"
    # subprocess_command = f"{sirius_path} --input {input_path} --output {output_path}.sirius config {config} formulas -p {instrument} fingerprints classes structures write-summaries"
    subprocess_command = f"{sirius_path} --input {input_path} --output '{output_path}.sirius' formulas fingerprints classes structures write-summaries"
    run_subprocess(subprocess_command)
    return True

def run_toxtree(input_path: str, output_path: str,toxtree_path: str, module_path: str) -> bool:
    """Runs a Toxtree module on infile from the command line

    Takes a file of SMILES as input to run toxtree from the command line. The specific toxtree module can be selected (see..). 
    Writes the results to fn_out. 

    Parameters
    ----------
    toxtree_path
        file path of toxtree jar file
    module path
        file path of toxtree module that is to be used 
    fn_in
        filename of file with SMILES strings on newlines with a "SMILES" header.
    fn_out
        output fn for toxtree classifications
        
    Returns
    -------
    bool
        True if function sucessfully ran, False otherwise
    
      """
    subprocess_command = f"java -jar {toxtree_path} -n m {module_path} -b {input_path} -i {input_path} -o {output_path}"
    cwd = Path(toxtree_path).parent
    run_subprocess(subprocess_command,cwd=cwd)
    return True
    # def run_toxtree(fn_out,fn_in):
    # toxtree_loc = "/lustre/BIF/nobackup/hendr218/Programs/Toxtree/Toxtree-v3.1.0.1851/Toxtree/Toxtree-3.1.0.1851.jar"
    # module_loc = "toxTree.tree.cramer.CramerRules"
 
    # subprocess.run([subprocess_command], cwd="/lustre/BIF/nobackup/hendr218/Programs/Toxtree/Toxtree-v3.1.0.1851/Toxtree", shell=True, capture_output=True, text=True)
    
    # df_sdf_cramer = rdkit.Chem.PandasTools.LoadSDF(fn_out)
#    # df_cramer = df_sdf_cramer.loc[:,['Cramer rules','SMILES']]
#     return fn_out

def run_classyfire(input_path: str,output_path: str) -> bool:
    """Runs classyfire for strings in infile, writes results to outfile
    
    Uses the classyfire API to classify compounds. First splits input file into batches of 999 compounds to avoid API restrictions.
    Waits between post request to avoid timeouts from the API, so can be slow. 

     Parameters
    ----------
    infile
        filename of file with SMILES strings on newlines
    outfile
        fn for output in .SDF format. 
        
    Returns
    -------
    bool
        True if function sucessfully ran, False otherwise
    
    """
    structures_batches  = file_to_batches(input_path, batchlength = 999)
    api_url: str = "http://classyfire.wishartlab.com" 
    retrieve_ids: list[str] = [post_to_classyfire(batch,api_url) for batch in structures_batches]
    classyfire_results = (get_classyfire_results(api_url,retrieve_id) for retrieve_id in retrieve_ids if check_classyfire_done(api_url, retrieve_id))
    with open (output_path,'a') as file: 
            file.writelines(classyfire_results)
    return True


def run_ms2lda(dataset,params):
    preprocessing_parameters,convergence_parameters,annotation_parameters,model_parameters,train_parameters, \
    fingerprint_parameters,dataset_parameters,n_iterations, n_motifs,motif_parameter = params
    
    MS2LDA.run(dataset, n_motifs=n_motifs, n_iterations=n_iterations,
    dataset_parameters=dataset_parameters,
    train_parameters=train_parameters,
    model_parameters=model_parameters,
    convergence_parameters=convergence_parameters,
    annotation_parameters=annotation_parameters,
    motif_parameter=motif_parameter,
    preprocessing_parameters=preprocessing_parameters,
    fingerprint_parameters=fingerprint_parameters)

    return True
  

def post_to_classyfire(structures_batch: str,base_url: str) -> str:
    """Makes a post request for classification to classyfire for structures (max 999)"""
    post_url: str = f"{base_url}/queries.json"
    data: str = '{"label": "myworkflow", "query_input": "'+structures_batch+'", "query_type": "STRUCTURE"}'
    headers: dict[str,str] = {"Content-Type": "application/json"}
    r = requests.post(post_url,data=data,headers=headers)
    retrieve_id: str = json.loads(r.text)['id']
    return retrieve_id

def check_classyfire_done(base_url: str,retrieve_id: str) -> bool:
    """Checks with API if batch is processed, waits untill it is to return True"""

#Get results from classyfire & create SDF
    status_url: str = f'{base_url}/queries/{retrieve_id}/status.json' 
    can_proceed: bool = False
    while not can_proceed:
        r = requests.get(status_url,headers={"accept": "json"})
        if r.text == "In progress":
            can_proceed = True
        elif (r.status_code != 200) | (r.text != "Done"):
            sleep(61)
        else:
            can_proceed = True
    return True
    # if requests.get(status_url,headers={"accept": "json"}).status_code == "504":
    #     sleep(60)
    # while requests.get(status_ur l,headers={"accept": "json"}).text != "Done":
    #     sleep(60) 
def get_classyfire_results(base_url: str,retrieve_id: str) -> str:
    """Obtains classyfire results in SDF for previsouly send batch via API"""
    retrieve_url: str = f'{base_url}/queries/{retrieve_id}.sdf'
    has_response: bool = False
    while not has_response:
        try:
            r = requests.get(retrieve_url,headers={"accept": "text/plain"},timeout=90)
            if r.status_code == 200:
                contents: str = r.text
                has_response = True
            else:
                sleep(60)
        except requests.exceptions.Timeout: # | (NameError):
            continue
    return contents

def set_ms2lda_params():
    preprocessing_parameters = {
        "min_mz": 0,
        "max_mz": 1000,
        "max_frags": 1000,
        "min_frags": 3,
        "min_intensity": 0.01,
        "max_intensity": 1
    }

    convergence_parameters = {
        "step_size": 50,
        "window_size": 10,
        "threshold": 0.001,
        "type": "perplexity_history"
    }
    annotation_parameters = {
        "criterium": "best", # return cluster with most compounds in it after optimization ("best" also an option)
        "cosine_similarity": 0.70, #0.8 how similar are the spectra compared to motifs in the optimization
        "n_mols_retrieved": 5 # 10 molecules retrieved from database by Spec2Vec
    }
    #Change if needed (cos score)


    model_parameters = {
        "rm_top": 0, #get rid of noisy point
        "min_cf": 0, #intensity within spectrum
        "min_df": 3, #min frequency within document
        "alpha": 0.6, #A higher alpha makes the document preferences "smoother" over topics
        "eta": 0.01, #and a higher eta makes the topic preferences "smoother" over words.
        "seed": 42,
    }

    train_parameters = {
        "parallel": 3,
        "workers": 1, 
    }

    fingerprint_parameters = {
    "fp_type": "rdkit",
    "threshold": 0.8,
    }

    dataset_parameters = {
        "acquisition_type": "DDA",
        "significant_digits": 2, #could be 3
        "charge": 1,
        "name": "pos_all",
        "output_folder": ""
    }
    return preprocessing_parameters,convergence_parameters,annotation_parameters,model_parameters,train_parameters,fingerprint_parameters,dataset_parameters
   



from metabomix import MetaboMix

def main():
    settings_path: str = "./Example_recipe.json"
    example_mix = MetaboMix(settings_path)
    example_mix.run_all()
    
if __name__ == "__main__":
    main()